
/*
 *	MASTER-MIND STREAM
 *
 *	Receives the input from libwww protocol modules and writes
 *	it to the client socket, doing various other tasks at the
 *	same time that used to be done in separate streams.
 *
 *	Small streams doing trivial tasks were really starting to
 *	effect the performance.  Streams are a fine way to add new
 *	features to software that is already complicated enough as
 *	it is without losing track of what is happening.  However,
 *	since the functionality has been put in successfully it is
 *	now time to rewrite it in an efficient and clean manner.
 *
 *	This stream handles:
 *	 * If-Modified-Since
 *	 * caching
 *	 * content-length counting
 *	 * HEAD
 *	 * CGI script responses
 *	 * file copy to socket (HTLoadFile() is no longer called for
 *	   others than directories)
 *
 * Author:
 *	Ari Luotonen, May 1994, luotonen@www.cern.ch
 */

#include "HTUtils.h"
#include "HTStream.h"
#include "HTSUtils.h"
#include "HTDaemon.h"
#include "HTConfig.h"


/************************************************************************
 *
 * MIME parser/bufferer for httpd
 *
 * This optionally parses either a full HTTP header including status
 * line, or just the MIME headers.
 *
 * This is used when reading a reply from a remote server when
 * proxying/caching (full HTTP), and also with CGI scripts (only MIME).
 *
 ************************************************************************/

/*
 * Continuation lines, just a list of strings.
 */
typedef struct _MIMECont {
    char *		line;
    struct _MIMECont *	next;
} MIMECont;

/*
 * MIME header without the status lines
 */
typedef struct _MIMEHeader {
    char *	name;		/* Header line name */
    char *	contents;	/* Header line value */
    MIMECont *	cont_head;	/* Continuation lines */
    MIMECont *	cont_tail;
    struct _MIMEHeader * next;
} MIMEHeader;

/*
 * HTTP response: the status line plus MIME headers
 */
typedef struct _HTTPHeader {
    int		header_length;	/* Calculated */
    BOOL	headers;	/* In header part, not in status line*/
    BOOL	eol;		/* End-of-line encountered */
    char *	cur_line;
    char *	cur_line_cur;
    char *	cur_line_end;
    int		status_code;	
    char *	status_line;
    MIMEHeader*	mime_head;
    MIMEHeader*	mime_tail;
} HTTPHeader;

PUBLIC HTTPHeader * HTTPHeader_new ARGS1(BOOL, full)
{
    HTTPHeader * h = (HTTPHeader*)calloc(1,sizeof(HTTPHeader));
    if (!full)
	h->headers = YES;	/* No status line, MIME headers right away */
    return h;
}

PRIVATE void increase_cur_line ARGS1(HTTPHeader *, h)
{
    int cur_size = h->cur_line_end - h->cur_line;
    h->cur_line = realloc(h->cur_line, 2*cur_size);
    h->cur_line_cur = h->cur_line + cur_size;
    h->cur_line_end = h->cur_line + 2*cur_size;
}

PRIVATE int get_status ARGS1(char *, buf)
{
    char * start = NULL;

    if (!buf || 0!=strncmp(buf, "HTTP/", 5) || !(start = strchr(buf, ' ')))
	return -1;

    while (*start && WHITE(*start)) start++;
    if (!*start) return -1;

    return atoi(start);
}

PRIVATE void hdr_cont ARGS2(MIMEHeader *,	hdr,
			    char *,		buf)
{
    MIMECont * cont;
    if (!hdr) { FREE(buf); return; }
    cont = (MIMECont*)calloc(1,sizeof(MIMECont));
    cont->line = buf;
    if (hdr->cont_tail)
	hdr->cont_tail->next = cont;
    else
	hdr->cont_head = cont;
    hdr->cont_tail = cont;
}

PRIVATE MIMEHeader * hdr_line ARGS1(char *, buf)
{
    char * colon = strchr(buf,':');
    MIMEHeader * m = (MIMEHeader*)calloc(1,sizeof(MIMEHeader));
    if (colon) {
	*colon++ = 0;
	while (*colon && WHITE(*colon)) colon++;
    }
    m->name = buf;
    m->contents = colon;
    return m;
}

#define LINE_LEN 85
PRIVATE void new_line ARGS1(HTTPHeader *, h)
{
    h->cur_line_cur = h->cur_line = (char*)malloc(LINE_LEN+1);
    h->cur_line_end = h->cur_line + LINE_LEN;
}

PRIVATE void process_line ARGS1(HTTPHeader *, h)
{
    *h->cur_line_cur = 0;

    if (h->headers) {	/* Normal MIME header lines */
	CTRACE(stderr,"Header line. %s\n", h->cur_line);
	if (WHITE(*h->cur_line))
	    hdr_cont(h->mime_tail,h->cur_line);
	else {
	    MIMEHeader * m = hdr_line(h->cur_line);
	    if (h->mime_tail)
		h->mime_tail->tail = m;
	    else
		h->mime_head = m;
	    h->mime_tail = m;
	}
    }
    else {	/* Status line */
	CTRACE(stderr, "Status line. %s\n", h->cur_line);
	h->status_code = get_status(h->cur_line);
	h->status_line = h->cur_line;
    }

    new_line(h);
}



/*
 * Parse HTTP header character by character.  Finds linebreaks;
 * returns finally YES when the header section is complete.
 *
 */
PRIVATE BOOL HTTPHeader_parse(HTTPHeader *,	h,
			      char,		c)
{
    h->header_length++;
    *h->cur_line_cur++ = c;
    if (h->cur_line_cur + 1 >= h->cur_line_end)
	increase_cur_line(h);

    if (!h->eol && c == LF) {
	me->eol = YES;
	process_line(h);
    }
    else if (h->eol) {
	if (c == LF) {
	    CTRACE(stderr, "Header...... complete\n");
	    return YES;
	}
	else if (c != CR) h->eol = NO;
    }
    return NO;
}




/************************************************************************
 *
 * MIME message body buffering
 *
 * This object just accepts blocks and buffers them.
 *
 ************************************************************************/

typedef struct _HTBufItem {
    int			len;
    char *		buf;
    struct _HTBufItem *	next;
} HTBufItem;

typedef struct _HTBodyBuffer {
    HTBufItem *	head;
    HTBufItem *	tail;
} HTBodyBuffer;


PRIVATE void append_buf ARGS3(HTStream *, me, char *, buf, int, len)
{
    HTBufItem * b = (HTBufItem*)calloc(1,sizeof(HTBufItem));
    if (!b) outofmem(__FILE__, "append_buf");

    b->len = len;
    b->buf = (char*)malloc(len);
    memcpy(b->buf,buf,len)

    if (me->buf_tail)
	me->buf_tail->next = b;
    else
	me->buf_head = b;
    me->buf_tail = b;
}

PRIVATE void buf_flush ARGS1(HTStream *, me)
{
    HTBufItem * b = NULL;

    if (me->do_send) {
	b = me->buf_head;
	while (b) {
	    NETWRITE(me->output_socket, b->buf, b->len);
	    b = b->next;
	}
    }
    b = me->buf_head;
    while (b) {
	HTBufItem * next = b->next;
	free(b->buf);
	free(b);
	b = next;
    }
    me->buf_head = me->buf_tail = NULL;
}

PRIVATE BOOL buf_alloc_new ARGS2(HTStream *,	me,
				 int,		size)
{
    if (me->content_count >= sc.max_content_len_buf) {
	CTRACE(stderr,
      "\nGiving up... buffering to find out content-length [limit %d bytes]\n",
	       sc.max_content_len_buf);
	out.content_length = 0;
	buf_flush(me);
	return NO;
    }
    else {
	me->tmp_ind = 0;
	me->tmp_max = size;
	me->tmp_buf = (char *)malloc(size);
	if (!me->tmp_buf) outofmem(__FILE__, "buf_put_char");
	return YES;
    }
}





struct _HTStream {
    HTStreamClass *	isa;
    HTStream *		sink;
    BOOL		body;		/* In body part */
    BOOL		do_send;	/* Send to client right away */
    BOOL		do_buffer;	/* Do buffering before sending */
    BOOL		do_ignore;	/* Ignore body completely */
    int			status_code;
    char *		status_line;
    MIMEHeader *	MIME_head;
    HIMEHeader *	MIME_tail;
    HTBufItem *		buf_head;
    HTBufItem *		buf_tail;
    char		tmp_buf[BUF_SIZE];
    int			tmp_buf_ind;
    int			content_length;
    int			output_socket;
    FILE *		output_cache;
};



PRIVATE char * reply_headers ARGS1(HTRequest *, req)
{
    return reply_headers_with(req, NULL, 0);
}

PUBLIC char * reply_headers_with ARGS3(Request *,	req,
				       char **,		extras,
				       int,		extra_cnt)
{
    char * headers = NULL;
    FILE * meta_file = NULL;
    int len = 1024;

    if (out.status_code <= 0) out.status_code = 500;
    if (!HTReasonLine) HTReasonLine = "Internal error";

    len += strlen(HTReasonLine);
    if (HTLocation) len += strlen(HTLocation);
    if (HTWWWAuthenticate) len += strlen(HTWWWAuthenticate);
    if (extras) {
	int i;
	for (i=0; i<extra_cnt; i++) {
	    if (extras[i])
		len += strlen(extras[i]) + 3;
	}
    }

    if (HTMetaFile) {
	struct stat stat_meta;

	if (stat(HTMetaFile, &stat_meta) != -1) {
	    len += stat_meta.st_size;
	    meta_file = fopen(HTMetaFile, "r");
	}
    }

    headers = (char*)malloc(len);
    if (!headers) outofmem(__FILE__, "HTReplyHeaders");

    sprintf(headers,
	    "%s %d %s\r\n%s\r\n%s%s",
	    HTServerProtocol, out.status_code, HTReasonLine,
	    "MIME-Version: 1.0",
	    "Server: CERN/", VD);

#ifndef VMS
    if (cur_time) {
	strcat(headers, "\r\nDate: ");
	strcat(headers, http_time(&cur_time));
    }
#endif	

    if (HTLocation) {		/* Do redirection */
	strcat(headers, "\r\nLocation: ");
	strcat(headers, HTLocation);
    }

    if (req->content_language) {
	strcat(headers, "\r\nContent-Language: ");
	strcat(headers, HTAtom_name(req->content_language));
    }

    if (req->content_encoding) {
	char * name = HTAtom_name(req->content_encoding);
	if (name && strcmp(name,"7bit") && strcmp(name,"8bit") &&
	    strcmp(name,"binary")) {
	    strcat(headers, "\r\nContent-Encoding: ");
	    strcat(headers, name);
	}
    }

    if (req->content_type) {
	strcat(headers, "\r\nContent-Type: ");
	strcat(headers, HTAtom_name(req->content_type));
    }

    if (out.content_length > 0) {
	char content_len[30];
	sprintf(content_len, "\r\nContent-Length: %d", out.content_length);
	strcat(headers, content_len);
    }

    if (HTLastModified) {
	strcat(headers, "\r\nLast-Modified: ");
	strcat(headers, HTLastModified);
    }

    if (HTExpires) {
	strcat(headers, "\r\nExpires: ");
	strcat(headers, HTExpires);
    }

    if (extras) {
	int i;
	for (i=0; i<extra_cnt; i++) {
	    if (extras[i] && *(extras[i])) {
		strcat(headers, "\r\n");
		strcat(headers, extras[i]);
	    }
	}
    }

    strcat(headers, "\r\n");
    if (HTWWWAuthenticate) {
	/*
	 * This is the last one, because it already contains \r\n
	 * in the end.
	 */
	strcat(headers, HTWWWAuthenticate);
    }

    if (meta_file) {
	char * buf = (char*)malloc(1001);
	if (!buf) outofmem(__FILE__, "HTReplyHeaders");

	while (fgets(buf, 1000, meta_file)) {
	    if (fix_meta_field(buf))
		strcat(headers, buf);
	}
	fclose(meta_file);
	free(buf);		/* Leak fixed AL 27 Feb 1994 */
    }

    strcat(headers, "\r\n");

    VTRACE(stderr,
	   "............ Header section for client\n%s%s",
	   headers, "............ End of headers\n");

    out.header_length = strlen(headers);
    CTRACE(stderr, "HTTP header. length: %d bytes\n", out.header_length);

    return headers;
}

PRIVATE void write_headers ARGS2(HTStream *,	me,
				 FILE *,	fp)
{
    MIMEHeader * hdr = me->MIME_head;

    fputs(me->status_line,fp);

    while (hdr) {
	MIMECont * cont = hdr->cont_head;
	fputs(hdr->name,fp);
	fputs(": ",fp);
	fputs(hdr->contents,fp);
	while (cont) {
	    fputs(cont->line);
	    cont = cont->next;
	}
	hdr = hdr->next;
    }
    fputs("\r\n",fp);
}

PRIVATE void MM_put_char ARGS2(HTStream *, me, char, c)
{
    if (me->do_ignore) return;
    if (me->body) {	/* This function should not be called when in body!! */
	me->content_length++;
	me->tmp_buf[me->tmp_buf_ind++] = c;
	if (me->tmp_buf_ind + 1 >= BUF_SIZE) {
	    if (me->do_send)
		NETWRITE(me->output_socket, me->tmp_buf, me->tmp_buf_ind);
	    else if (me->do_buffer)
		buf_append(me, me->tmp_buf, me->tmp_buf_ind);
	    me->tmp_buf_ind = 0;
	}
	return;
    }

.............
}

PRIVATE void MM_put_block ARGS3(HTStream *, me, CONST char *, b, int, l)
{
    if (me->do_ignore) return;
    while (!me->body && l-- > 0)
	MM_put_char(me,*b++);
    if (l <= 0) return;

    me->content_length += l;

    if (me->do_send) {
	if (me->tmp_buf_ind > 0) {
	    NETWRITE(me->output_socket, me->tmp_buf, me->tmp_buf_ind);
	    me->tmp_buf_ind = 0;
	}
	NETWRITE(me->output_socket, b, l);
    }
    else if (me->do_buffer) {
	if (me->tmp_buf_ind) {
	    buf_append(me, me->tmp_buf, me->tmp_buf_ind);
	    me->tmp_buf_ind = 0;
	}
	buf_append(me,b,l);
    }
}

PRIVATE void MM_put_string ARGS2(HTStream *, me, CONST char*, s)
{
    int len = s ? strlen(s) : 0;
    if (len > 0) MM_put_block(me,s,len);
}

PRIVATE void MM_free ARGS1(HTStream *, me)
{
}

PRIVATE void MM_abort ARGS2(HTStream *, me, HTError, e)
{
}

